import 'dart:convert';
import 'dart:io';
import 'package:dio/adapter.dart';
import 'package:dio/dio.dart';
import 'package:dio_logger/dio_logger.dart';
import 'package:easyflu/config/const.dart';
import 'package:easyflu/utils/object_util.dart';
import 'exception/http_error.dart';
import 'http_data.dart';


/// 请求方法.
class Method {
  static final String get = "GET";
  static final String post = "POST";
  static final String put = "PUT";
  static final String head = "HEAD";
  static final String delete = "DELETE";
  static final String patch = "PATCH";
}


class EasyDio {

  //服务器路径
  static EasyDio _instance = EasyDio._internal();

  late Dio _dio;

  factory EasyDio() => _instance;

  late BaseOptions _baseOptions;

  Function? _onError;

  Function? _onFinish;

  Function? _onStart;

  ValidateResponse? _validateResponse;

  late String _successCode;

  late String _codeKey;

  late String _msgKey;

  late String _dataKey;

  late bool _checkDataEmpty;

  late bool openLog;

  late String? pem;

  late int retry;

  late String? pKCSPath;

  late String? pKCSPwd;

  late List<Interceptor>? interceptors;

  Dio get dio => _dio;

  List<String> publicErrorCode = [];

  ///-----------------------------------------工具类初始化设置-------------------------------------

  EasyDio._internal();


  ///初始化公共属性
  ///
  /// [baseUrl] 地址前缀
  /// [connectTimeout] 连接超时时间，毫秒
  /// [receiveTimeout] 接收超时时间，毫秒
  /// [interceptors] 基础拦截器
  void init({String baseUrl = "",
    int connectTimeout = 30 * 1000,
    int receiveTimeout = 30 * 1000,
    String? pem,
    String? pKCSPath,
    String? pKCSPwd,
    String codeKey = '',
    String msgKey = '',
    String dataKey = '',
    String successCode = '',
    bool checkDataEmpty = true,
    ResponseType? responseType,
    bool openLog = true,
    int retry = 3,
    String? method,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? extra,
    String? contentType,
    ValidateStatus? validateStatus,
    bool? followRedirects,
    Function? onError,
    Function? onFinish,
    Function? onStart,
    ValidateResponse? validateResponse,
    bool enableAutoCookie = false,
    ListFormat? listFormat,
    List<String>? publicErrorCode,
    List<Interceptor>? interceptors
  }) {

    this._validateResponse = validateResponse;
    this._onStart = onStart;
    this._onError = onError;
    this._onFinish = onFinish;
    this._successCode = successCode;
    this._codeKey = codeKey;
    this._msgKey = msgKey;
    this._dataKey = dataKey;
    this._checkDataEmpty = checkDataEmpty;
    this.openLog = openLog;
    this.retry = retry;
    this.pem = pem;
    this.pKCSPath = pKCSPath;
    this.interceptors = interceptors;


    method = method??Method.post;

    if(publicErrorCode != null){
      this.publicErrorCode.addAll(publicErrorCode);
    }

    _baseOptions = BaseOptions(
        baseUrl: baseUrl,
        receiveTimeout: receiveTimeout,
        connectTimeout: connectTimeout,
        contentType: contentType,
        responseType: responseType??ResponseType.json,
        headers: headers??{},
        extra: extra??{},
        validateStatus: validateStatus,
        followRedirects: followRedirects,
        listFormat: listFormat,
        method: method
    );

    _dio = createDio();

    if(interceptors != null){
      _dio.interceptors.addAll(interceptors);
    }
  }


  ///设置BaseOptions
  BaseOptions copyBaseOptions(){
    return BaseOptions(
        baseUrl: _baseOptions.baseUrl,
        receiveTimeout: _baseOptions.receiveTimeout,
        connectTimeout: _baseOptions.connectTimeout,
        contentType: _baseOptions.contentType,
        responseType: _baseOptions.responseType,
        headers: _baseOptions.headers,
        extra: _baseOptions.extra,
        validateStatus: _baseOptions.validateStatus,
        followRedirects: _baseOptions.followRedirects,
        listFormat: _baseOptions.listFormat,
        method: _baseOptions.method
    );
  }

  Dio createDio(){
    Dio dio = Dio(copyBaseOptions());

    if(openLog && _isInDebugMode){
      dio.interceptors.add(dioLoggerInterceptor);
    }


    if (retry != 0) {
      // dio.interceptors.add(element)
      // dio.interceptors.add(RetryInterceptor(
      //     dio: dio,
      //     options: RetryOptions(
      //       retries: retry, // Number of retries before a failure
      //       retryInterval: const Duration(seconds: 1), // Interval between each retry
      //     )
      // ));
    }



    ///-----------------------------------------证书相关信息-------------------------------------
    (dio.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate = (HttpClient client) {
      client.badCertificateCallback = (X509Certificate cert, String host, int port) {
        if (pem != null) {
          if (cert.pem == pem) {
            // 证书一致，则放行
            return true;
          }else{
            return false;
          }
        }else{
          return true;
        }
      };
    };

    if (pKCSPath != null && pKCSPwd != null) {
      (dio.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate = (HttpClient client) {
        SecurityContext sc = new SecurityContext();
        //file为证书路径
        sc.setTrustedCertificates(pKCSPath!, password: pKCSPwd);
        HttpClient httpClient = new HttpClient(context: sc);
        return httpClient;
      };
    }

    return dio;
  }


  bool get _isInDebugMode {
    bool inDebugMode = false;
    assert(inDebugMode = true);
    return inDebugMode;
  }


  ///-----------------------------------------请求方法封装-------------------------------------
  Future<HttpData<T>> request<T>({
    String? url,
    String? method,
    Options? options,
    String? cacheKey,
    bool? isFormData,
    bool? publicMethod,
    var data,
    String? contentType,
    CancelToken? cancelToken,
    Function? onStart,
    Function(HttpData<T> httpData)? onSuccess,
    Function(HttpData httpData)? onError,
    Function? onFinish,
    ValidateResponse? validateResponse,
    bool returnOriginData = false,
    String? successCode,
    String? codeKey,
    String? msgKey,
    String? dataKey,
    bool? checkDataEmpty,
    ProgressCallback? onReceiveProgress,
    ProgressCallback? onSendProgress,
    dynamic Function(dynamic map)? handleRes,
    HttpError Function(dynamic data)? validateError,
    bool useNewDio = false,
  }) async{

    return await _handleHttpData<T, dynamic>(
        url: url,
        method: method,
        options: options,
        cacheKey: cacheKey,
        isFormData: isFormData,
        publicMethod: publicMethod,
        data: data,
        cancelToken: cancelToken,
        onStart: onStart,
        onSuccess: onSuccess,
        onError: onError,
        onFinish: onFinish,
        validateResponse: validateResponse,
        validateError: validateError,
        successCode: successCode,
        codeKey: codeKey,
        msgKey: msgKey,
        dataKey: dataKey,
        onReceiveProgress: onReceiveProgress,
        onSendProgress: onSendProgress,
        handleRes: handleRes,
        checkDataEmpty: checkDataEmpty,
        useNewDio: useNewDio
    );
  }


  Future<HttpData<T>> requestList<T, E>({
    String? url,
    String? method,
    Options? options,
    String? cacheKey,
    bool? isFormData,
    bool? publicMethod,
    var data,
    String? contentType,
    CancelToken? cancelToken,
    Function? onStart,
    Function(HttpData<T> httpData)? onSuccess,
    Function(HttpData httpData)? onError,
    Function? onFinish,
    ValidateResponse? validateResponse,
    String? successCode,
    String? codeKey,
    String? msgKey,
    String? dataKey,
    bool? checkDataEmpty,
    ProgressCallback? onReceiveProgress,
    ProgressCallback? onSendProgress,
    dynamic Function(dynamic map)? handleRes,
    HttpError Function(dynamic data)? validateError,
    bool useNewDio = false,
  }) async{


    return await _handleHttpData<T, E>(
        url: url,
        method: method,
        options: options,
        cacheKey: cacheKey,
        isFormData: isFormData,
        publicMethod: publicMethod,
        data: data,
        cancelToken: cancelToken,
        onStart: onStart,
        onSuccess: onSuccess,
        onError: onError,
        onFinish: onFinish,
        validateResponse: validateResponse,
        validateError: validateError,
        successCode: successCode,
        codeKey: codeKey,
        msgKey: msgKey,
        dataKey: dataKey,
        onReceiveProgress: onReceiveProgress,
        onSendProgress: onSendProgress,
        handleRes: handleRes,
        checkDataEmpty: checkDataEmpty,
        useNewDio: useNewDio
    );
  }


  ///进行httpData组装
  Future<HttpData<T>> _handleHttpData<T, E>({
    String? url,
    String? method,
    Options? options,
    String? cacheKey,
    bool? isFormData,
    bool? publicMethod,
    var data,
    CancelToken? cancelToken,
    Function? onStart,
    Function(HttpData<T> httpData)? onSuccess,
    Function(HttpData httpData)? onError,
    Function? onFinish,
    ValidateResponse? validateResponse,
    HttpError Function(dynamic data)? validateError,
    String? successCode,
    String? codeKey,
    String? msgKey,
    bool? checkDataEmpty,
    String? dataKey,
    ProgressCallback? onReceiveProgress,
    ProgressCallback? onSendProgress,
    dynamic Function(Map map)? handleRes,
    bool useNewDio = false,
  }) async{

    Dio dio = _dio;

    if(useNewDio){
      dio = createDio();
    }

    HttpData<T> httpData = HttpData<T>();

    ///httpData组装数据
    String code = "";
    String? msg;
    dynamic resultData;


    bool mIsFormData = isFormData??false;
    bool mPublicMethod = publicMethod??true;

    try{

      if(onStart != null){
        onStart.call();
      }else if(mPublicMethod) {
        this._onStart?.call();
      }


      Response response;

      Options mOptions = Options();

      if(options != null){
        mOptions = options;
      }

      if (mIsFormData) {
        data = FormData.fromMap(data);
      }

      if(method != null){
        mOptions.method =  method;
      }else{
        mOptions.method = _baseOptions.method;
      }

      if(url == null){
        throw HttpError.urlError(url);
      }



      if (method == Method.get) {
        response = await dio.request(
            url,
            queryParameters: data,
            options: mOptions,
            cancelToken: cancelToken,
            onReceiveProgress: (int sent, int total){
              onReceiveProgress?.call(sent, total);
            },
            onSendProgress: (int sent, int total){
              onSendProgress?.call(sent, total);
            }
        );
      }else{
        response = await dio.request(
            url,
            data: data,
            options: mOptions,
            cancelToken: cancelToken,
            onReceiveProgress: (int sent, int total){
              onReceiveProgress?.call(sent, total);
            },
            onSendProgress: (int sent, int total){
              onSendProgress?.call(sent, total);
            }
        );
      }


      String mCodeKey = codeKey??_codeKey;
      String mMsgKey = msgKey??_msgKey;
      String mDataKey = dataKey??_dataKey;
      String mSuccessCode = successCode??_successCode;
      checkDataEmpty = checkDataEmpty ?? _checkDataEmpty;


      if (response.data is List) {
        resultData = response.data;
        code = successCode??_successCode;
      }else{
        Map<String, dynamic> resJson;

        httpData.rawData = response.data;

        try{
          if(response.data is Map){
            resJson = response.data;
          }else{
            resJson = json.decode(response.data);
          }
        }catch(e){
          throw HttpError.parseDataError(e, logMsg: 'response.data转换json报错');
        }

        if(ObjectUtil.isNotEmpty(mCodeKey) && resJson.containsKey(mCodeKey)){
          code = resJson[mCodeKey].toString();
        }

        if(ObjectUtil.isNotEmpty(mMsgKey) && resJson.containsKey(mMsgKey)){
          msg = resJson[mMsgKey].toString();
        }



        if (validateResponse != null) {
          resJson = validateResponse(response.data);
        }else if(_validateResponse != null){
          resJson = _validateResponse!(response.data);
        }


        if(code != mSuccessCode){
          throw HttpError.codeError(mSuccessCode, code, msg: msg);
        }else{

          ///解析data
          var rawData;
          if(ObjectUtil.isEmpty(mDataKey)){
            rawData = resJson;
          }else{
            rawData = resJson[mDataKey];
          }


          if(handleRes != null){
            try{
              if(rawData is List){
                if(ObjectUtil.isNotEmpty(rawData)){
                  if(rawData.first is Map){
                    resultData = List<E>.from(rawData.map((e) => handleRes(e)).toList());
                  }else{
                    resultData = rawData;
                  }
                }
              }else{
                resultData = handleRes(rawData);
              }
            }catch(e){
              throw HttpError.parseDataError(e, logMsg: "在解析data时出现错误");
            }
          }else{
            resultData = rawData;
          }

          if(ObjectUtil.isEmpty(resultData) && checkDataEmpty){
            throw HttpError.emptyDataError(response.data);
          }
        }
      }
    }catch(e){
      HttpError error;
      if (e is DioError) {
        error = HttpError.dioError(e);
      }else if(e is HttpError){
        error = e;
      }else{
        error = HttpError.unknown(e);
      }
      httpData.httpError = error;
    }




    ///开始组装httpData
    httpData.code = code;
    httpData.msg = msg;

    if(resultData is T){
      httpData.data = resultData;
    }


    try{
      if(httpData.httpError != null){
        if(validateError != null){
          httpData.httpError = validateError(httpData.data);
        }

        if(onError != null){
          onError.call(httpData);
          if(publicErrorCode.contains(code)){
            if (this._onError != null) {
              this._onError!.call(httpData);
            }
          }
        }else if(publicMethod ?? true) {
          this._onError?.call(httpData);
        }
      }else{
        onSuccess?.call(httpData);
      }


      if(onFinish != null){
        onFinish.call();
      }else if(mPublicMethod) {
        this._onFinish?.call();
      }

    }catch(e){
      httpData.httpError = HttpError.unknown(e, logMsg: '在组装httpData出现未知错误');
    }

    return httpData;
  }





  ///-----------------------------------------上传文件方法封装-------------------------------------------------------------
  Future<HttpData> fileUpload({
    MultipartFile? multipartFile,
    String? url,
    String filePath = "",
    String key = 'file',
    String? fileName,
    String? method,
    FormData? formData,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
    Function? onStart,
    Function? onError,
    Function? onFinish,
    dynamic Function(dynamic map)? handleRes,
    bool publicMethod = false
  }) async{

    MultipartFile file = multipartFile??await MultipartFile.fromFile(filePath, filename: fileName);

    return await request(
        url: url,
        method: method ?? Method.post,
        publicMethod: publicMethod,
        data: formData ?? FormData.fromMap({
          key: file
        }),
        onSendProgress: (int sent, int total){
          onSendProgress?.call(sent, total);
        },
        onReceiveProgress: (int sent, int total){
          onReceiveProgress?.call(sent, total);
        },
        handleRes: handleRes
    );

  }


  Future<HttpData> filesUpload({
    List<MultipartFile>? multipartFile,
    String? url,
    String? filePath,
    String key = 'file',
    String? fileName,
    String? method,
    FormData? formData,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
    Function? onStart,
    Function? onError,
    Function? onFinish,
    bool publicMethod = false
  }) async{


    return await request(
      url: url,
      method: method ?? Method.post,
      publicMethod: publicMethod,
      data: formData??FormData.fromMap({
        key: multipartFile
      }),
      onSendProgress: (int sent, int total){
        onSendProgress?.call(sent, total);
      },
      onReceiveProgress: (int sent, int total){
        onReceiveProgress?.call(sent, total);
      },
    );
  }


  ///-----------------------------------------下载文件方法封装-------------------------------------------------------------
  void download({
    String urlPath = '',
    String savePath = '',
    bool deleteOnError = true,
    String lengthHeader = Headers.contentLengthHeader,
    ProgressCallback? onReceiveProgress,
    data,
    Map<String, dynamic>? queryParameters,
    CancelToken? cancelToken,
    Options? options,
    Function? onStart,
    Function? onError,
    Function? onFinish,
    bool publicMethod = true
  }) async{

    try{
      await _dio.download(
        urlPath,
        savePath,
        deleteOnError: deleteOnError,
        lengthHeader: lengthHeader,
        data: data,
        options: options,
        queryParameters: queryParameters,
        cancelToken: cancelToken,
        onReceiveProgress: (int sent, int total){
          onReceiveProgress?.call(sent, total);
        },
      );
    }on DioError catch(e){
      throw HttpError.dioError(e);
    }catch(e){
      if (e is HttpError) {
        if(onError != null){
          onError.call(e);
        }else if (publicMethod) {
          this._onError?.call(e);
        }

        if (onFinish != null) {
          onFinish.call();
        } else if(publicMethod){
          this._onFinish?.call();
        }
      }
    }
  }




  ///-----------------------------------------dio部分设置-------------------------------------


  void setCookie(String cookie){
    Map<String, dynamic> _headers = new Map();
    _headers["Cookie"] = cookie;
    _dio.options.headers.addAll(_headers);

  }


  void setHeaders(Map<String, dynamic> headers){
    _baseOptions.headers.addAll(headers);
    _dio.options = _baseOptions;
  }

  void removeHeaders(String key){
    _baseOptions.headers.remove(key);
    _dio.options = _baseOptions;
  }

  ///修改默认url
  void changeBaseUrl(String baseUrl){
    _baseOptions.baseUrl = baseUrl;
    _dio.options = _baseOptions;
  }

}