import 'package:easyflu/config/const.dart';
import 'package:easyflu/easyflu.dart';
import 'package:easyflu/utils/win_media.dart';
import 'package:flutter/material.dart';
import 'dart:math' as math;


bool _hasDivider({double dividerThickness = 0, IndexedWidgetBuilder? separatorBuilder}){
  return dividerThickness != 0 || separatorBuilder != null;
}



///适应用于Sliver的分割线
SliverChildBuilderDelegate _getSeparatorSliverChildBuilder({
  required int itemCount,
  required IndexedWidgetBuilder itemBuilder,
  required IndexedWidgetBuilder separatorBuilder,
}){
  return SliverChildBuilderDelegate((BuildContext context, int index) {
    final int itemIndex = index ~/ 2;
    final Widget widget;
    if (index.isEven) {
      widget = itemBuilder(context, itemIndex);
    }else {
      widget = separatorBuilder(context, itemIndex);
    }
    return widget;
  },
    childCount: math.max(0, itemCount * 2 - 1),
    semanticIndexCallback: (Widget _, int index) {
      return index.isEven ? index ~/ 2 : null;
    },
  );
}


SliverChildDelegate _getSliverChildDelegate({
  ///item渲染器
  required IndexedWidgetBuilder itemBuilder,
  ///数据个数
  required itemCount,
  ///分割线高度
  double dividerThickness = 0,
  ///分割线颜色
  Color dividerColor = Colors.transparent,
  ///是否包含开头分割线
  bool dividerEnd = false,
  ///是否包含结尾分割线
  bool dividerStart = false,
  ///滑动方向
  Axis scrollDirection = Axis.vertical,
  ///分割线渲染器
  IndexedWidgetBuilder? separatorBuilder,
}){

  ///需要判断是否需要添加头部尾部分割线
  bool hasDivider = _hasDivider(
    dividerThickness: dividerThickness,
    separatorBuilder: separatorBuilder
  );

  SliverChildDelegate delegate;

  if(hasDivider){
    int realItemCount = itemCount;

    if(dividerStart){
      realItemCount = itemCount + 1;
    }

    if(dividerEnd){
      realItemCount = itemCount + 1;
    }

    delegate = _getSeparatorSliverChildBuilder(
        itemBuilder: (context, index){
          if(realItemCount == itemCount){
            return itemBuilder.call(context, index);
          }else{
            if(dividerStart && dividerEnd){
              if(index == 0 || index == realItemCount - 1){
                return Container();
              }else{
                return itemBuilder.call(context, index - 1);
              }
            }else{
              if(dividerStart){
                if(index == 0){
                  return Container();
                }else{
                  return itemBuilder.call(context, index);
                }
              }else{
                if(index == realItemCount - 1){
                  return Container();
                }else{
                  return itemBuilder.call(context, index);
                }
              }
            }
          }
        },
        separatorBuilder: separatorBuilder ?? (context, index){
          ///渲染分割线逻辑
          if(scrollDirection == Axis.vertical){
            return Container(
              height: getHeight(dividerThickness),
              color: dividerColor,
            );
          }else{
            return Container(
              width: getWidth(dividerThickness),
              color: dividerColor,
            );
          }
        },
        itemCount: realItemCount
    );
  }else{
    delegate = SliverChildBuilderDelegate(itemBuilder, childCount: itemCount);
  }
  return delegate;
}





Widget SmartSliverListView({
  Key? key,
  ///item渲染器
  required IndexedWidgetBuilder itemBuilder,
  ///数据个数
  required itemCount,
  ///分割线高度
  double dividerThickness = 0,
  ///分割线颜色
  Color dividerColor = Colors.transparent,
  ///是否包含开头分割线
  bool dividerEnd = false,
  ///是否包含结尾分割线
  bool dividerStart = false,
  ///滑动方向
  Axis scrollDirection = Axis.vertical,
  ///分割线渲染器
  IndexedWidgetBuilder? separatorBuilder,
  ///滑动模式
  ScrollPhysics? physics,
  ///列表元素回收
  CollectGarbage? collectGarbage,
  ///追踪进入 Viewport 的列表元素的 index（即你看到的可视区域，并不包括缓存距离）
  ViewPortBuilder? viewportBuilder,
}){
  bool hasDivider = _hasDivider(
      dividerThickness: dividerThickness,
      separatorBuilder: separatorBuilder
  );

  return ExtendedSliverList(
    delegate: _getSliverChildDelegate(
        itemBuilder: itemBuilder,
        separatorBuilder: separatorBuilder,
        scrollDirection: scrollDirection,
        dividerStart: dividerStart,
        dividerEnd: dividerEnd,
        dividerColor: dividerColor,
        dividerThickness: dividerThickness,
        itemCount: itemCount
    ),
    extendedListDelegate: ExtendedListDelegate(
      collectGarbage: collectGarbage,
      viewportBuilder: (first, last){
        if(hasDivider){
          viewportBuilder?.call(first ~/ 2, last ~/ 2);
        }else{
          viewportBuilder?.call(first, last);
        }
      }
    ),
  );
}



ScrollView SmartListView({
  Key? key,
  ///item渲染器
  required IndexedWidgetBuilder itemBuilder,
  ///数据个数
  required itemCount,
  ///分割线高度
  double dividerThickness = 0,
  ///分割线颜色
  Color dividerColor = Colors.transparent,
  ///缓存高度
  double? cacheExtent,
  ///是否包含开头分割线
  bool dividerEnd = false,
  ///是否包含结尾分割线
  bool dividerStart = false,
  ///滑动方向
  Axis scrollDirection = Axis.vertical,
  ///分割线渲染器
  IndexedWidgetBuilder? separatorBuilder,
  ///滑动模式
  ScrollPhysics? physics,
  ///列表元素回收
  CollectGarbage? collectGarbage,
  ///追踪进入 Viewport 的列表元素的 index（即你看到的可视区域，并不包括缓存距离）
  ViewPortBuilder? viewportBuilder,
  ///是否计算item尺寸
  bool shrinkWrap = false,
  ///是否使用Column
  bool useListView = false,
  ///滑动控制器
  ScrollController? scrollController,
}){

  bool hasDivider = _hasDivider(
      dividerThickness: dividerThickness,
      separatorBuilder: separatorBuilder
  );

  ///无分割线，使用普通的listview
  return ExtendedListView.custom(
    shrinkWrap: shrinkWrap,
    cacheExtent: cacheExtent,
    scrollDirection: scrollDirection,
    controller: scrollController,
    childrenDelegate: _getSliverChildDelegate(
        itemBuilder: itemBuilder,
        separatorBuilder: separatorBuilder,
        scrollDirection: scrollDirection,
        dividerStart: dividerStart,
        dividerEnd: dividerEnd,
        dividerColor: dividerColor,
        dividerThickness: dividerThickness,
        itemCount: itemCount
    ),
    extendedListDelegate: ExtendedListDelegate(
        collectGarbage: collectGarbage,
        viewportBuilder: (first, last){
          if(hasDivider){
            viewportBuilder?.call(first ~/ 2, last ~/ 2);
          }else{
            viewportBuilder?.call(first, last);
          }
        }
    ),
  );


}

